# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Neural networks for guiding heuristic search.
"""

from slidingpuzzle.nn.dataset import (
    SlidingPuzzleDataset,
    load_or_build_dataset,
    load_examples,
    make_examples,
    save_examples,
)

from slidingpuzzle.nn.eval import (
    accuracy,
    evaluate,
    eval_checkpoint,
)

from slidingpuzzle.nn.heuristics import (
    get_heuristic,
    make_heuristic,
    set_heuristic,
    v1_distance,
)

from slidingpuzzle.nn.models import (
    Model_v1,
    load_model,
    save_model,
)

from slidingpuzzle.nn.paths import (
    CHECKPOINT_DIR,
    DATASET_DIR,
    TENSORBOARD_DIR,
    get_checkpoint_path,
    get_examples_path,
    get_log_dir,
)

from slidingpuzzle.nn.train import (
    launch_tensorboard,
    load_checkpoint,
    save_checkpoint,
    set_seed,
    train,
)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A package for solving sliding tile puzzles.

Examples:
    >>> from slidingpuzzle import *
    >>> b = new_board(3, 3)
    >>> print_board(b)
    1 2 3
    4 5 6
    7 8
    >>> print_board(shuffle_board(b))
    1 6 7
    4   8
    5 3 2
    >>> search(b, "greedy", heuristic=manhattan_distance)
    solution=[3, 2, 8, 3, 6, 7, 3, 6, 7, 1, 4, 7, 2, 5, 7, 4, 1, 2, 5, 8]
    solution_len=20, generated=829, expanded=518, unvisited=312, visited=313
"""

from slidingpuzzle.heuristics import (
    corner_tiles_distance,
    euclidean_distance,
    hamming_distance,
    last_moves_distance,
    linear_conflict_distance,
    manhattan_distance,
    random_distance,
    relaxed_adjacency_distance,
)

from slidingpuzzle.board import (
    BLANK_TILE,
    board_from,
    board_from_iter,
    board_generator,
    count_inversions,
    find_blank,
    find_tile,
    flatten_board,
    freeze_board,
    get_goal_tile,
    get_goal_x,
    get_goal_y,
    get_goal_yx,
    get_next_moves,
    is_solvable,
    is_solved,
    new_board,
    print_board,
    random_move,
    shuffle_board,
    shuffle_board_lazy,
    solution_as_tiles,
)

from slidingpuzzle.algorithms import (
    ALGORITHMS,
    A_STAR,
    BEAM,
    BFS,
    DFS,
    GREEDY,
    IDA_STAR,
    IDDFS,
    compare,
    evaluate,
    get_next_states,
    search,
    swap_tiles,
)

from slidingpuzzle.state import SearchResult, State

import logging


logging.basicConfig(level=logging.INFO)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from slidingpuzzle import *


def test_new_board():
    ground_truth = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 0]])
    assert np.array_equal(new_board(3, 3), ground_truth)

    ground_truth = np.array([[1, 2], [3, 4], [5, 0]])
    assert np.array_equal(new_board(3, 2), ground_truth)


def test_board_from():
    assert np.array_equal(board_from([1, 2], [3, 0]), new_board(2, 2))


@pytest.mark.parametrize("h", [3, 5])
@pytest.mark.parametrize("w", [3, 5])
def test_board_from_iter(h, w):
    values = list(range(1, 1 + h * w))
    values[-1] = BLANK_TILE
    b = board_from_iter(h, w, values)
    assert np.array_equal(b, new_board(h, w))


def test_freeze_board():
    b = new_board(3, 3)
    frozen = freeze_board(b)
    assert id(b) != id(frozen)
    swap_tiles(b, (0, 1), (1, 0))
    assert b[0, 1] != frozen[0][1]


def test_print_board():
    print_board(new_board(5, 5))


def test_get_goal_yx():
    h, w = 3, 3
    board = new_board(h, w)
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            assert (y, x) == get_goal_yx(h, w, tile)


def test_find_blank():
    board = new_board(3, 3)
    assert find_blank(board) == (2, 2)
    assert find_tile(board, BLANK_TILE) == find_blank(board)

    board[0, 1], board[2, 2] = board[2, 2], board[0, 1]
    assert find_blank(board) == (0, 1)


def test_swap_tiles():
    board = new_board(3, 3)
    pos1 = (2, 2)
    pos2 = (1, 1)
    orig_pos1 = board[pos1[0], pos1[1]]
    orig_pos2 = board[pos2[0], pos2[1]]
    swap_tiles(board, pos1, pos2)
    assert orig_pos1 == board[pos2[0], pos2[1]]
    assert orig_pos2 == board[pos1[0], pos1[1]]


@pytest.mark.parametrize("size", [(2, 2), (3, 2), (2, 3), (3, 3)])
def test_shuffle_board(size):
    h, w = size
    board = new_board(h, w)
    shuffle_board(board)
    r = search(board, alg="greedy", heuristic=linear_conflict_distance)
    assert r.solution is not None


@pytest.mark.parametrize("size", [(2, 2), (3, 2), (2, 3), (3, 3)])
def test_shuffle_board_lazy(size):
    h, w = size
    board = new_board(h, w)
    shuffle_board_lazy(board)
    r = search(board, alg="greedy", heuristic=manhattan_distance)
    assert r.solution is not None


def test_get_possible_moves():
    board = new_board(3, 3)
    moves = get_next_moves(board, (2, 2))
    assert moves == [(2, 1), (1, 2)]

    pos1 = (1, 1)
    pos2 = (2, 2)
    swap_tiles(board, pos1, pos2)
    moves = get_next_moves(board, pos1)
    assert len(moves) == 4


def test_get_next_states():
    board = new_board(3, 3)
    state = State(board, (2, 2), [])
    next_states = get_next_states(state)
    assert len(next_states) == 2


@pytest.mark.parametrize("h", [3, 5])
@pytest.mark.parametrize("w", [3, 5])
def test_board_generator(h, w):
    gen = board_generator(h, w)
    assert np.array_equal(next(gen), board_from_iter(h, w, range(h * w)))



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines a PyTorch model to evaluate sliding puzzle boards.
"""

from typing import Optional

import math

import torch
import torch.nn as nn

import slidingpuzzle.nn.paths as paths
import slidingpuzzle.board as board_


VERSION_1 = "v1"


class Model_v1(nn.Module):
    """
    A stack of linear layers that accepts a board as input and outputs the
    estimated distance to the goal.
    """

    def __init__(self, h: int, w: int) -> None:
        super().__init__()
        self.version = VERSION_1  # required
        self.h = h  # required
        self.w = w  # required
        size = h * w
        self.flatten = nn.Flatten()
        self.linear1 = nn.Linear(size, size * 8, dtype=torch.float32)
        self.linear2 = nn.Linear(size * 8, size * 8, dtype=torch.float32)
        self.linear3 = nn.Linear(size * 8, size * 4, dtype=torch.float32)
        self.linear4 = nn.Linear(size * 4, size, dtype=torch.float32)
        self.linear5 = nn.Linear(size, 1, dtype=torch.float32)

    def forward(self, x):
        x = self.flatten(x)
        x = torch.relu(self.linear1(x))
        x = torch.relu(self.linear2(x))
        x = torch.relu(self.linear3(x))
        x = torch.relu(self.linear4(x))
        x = self.linear5(x)
        return x


def save_model(model: nn.Module, device: Optional[str] = None) -> None:
    """
    Save a frozen version of the model into the "models" dir.

    Args:
        model: The trained model.
        device: The device the model is currently loaded on. If not provied, it will
            be guessed.
    """
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    model.eval()
    model.to(device)
    path = paths.get_model_path(model.h, model.w, model.version)
    board = board_.new_board(model.h, model.w)
    example_inputs = torch.tensor(board, dtype=torch.float32).unsqueeze(0).to(device)
    traced_model = torch.jit.trace(model, example_inputs)
    frozen_model = torch.jit.freeze(traced_model)
    frozen_model.save(str(path))


def load_model(
    h: int, w: int, version: str, device: Optional[str] = None
) -> torch.ScriptModule:
    """
    Reload a pre-trained frozen model.
    """
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    model_path = paths.get_model_path(h, w, version)
    model = torch.jit.load(str(model_path), map_location=device)
    model.eval()
    return model



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module used to evaluate model performance vs. other heuristics.
"""
from typing import Optional

import torch
import torch.utils


def accuracy(expected, predicted) -> float:
    r"""
    Helper function to estimate accuracy of sliding puzzle model outputs.
    This function replaces NaNs/infinities with 0s and uses:

    .. math::

        f(e, p) = 1 - \frac{|e - p|}{1 + |e - p|}

    Args:
        expected: A tensor with the expected value
        predicted: A tensor with the predicted value

    Returns:
        The accuracy as a float in the range [0, 1].
    """
    diff = torch.abs(expected - predicted)
    diff /= 1 + diff
    diff = torch.nan_to_num(diff, nan=0, posinf=0, neginf=0)
    diff = 1 - diff
    return torch.sum(diff).item()


def evaluate(
    model: torch.nn.Module, criterion, dataset: torch.utils.data.Dataset
) -> tuple[float, float]:
    """
    Runs the model on provided datatset computing average loss and accuracy.

    Args:
        model: Model to evaluate
        criterion: The criterion function to use for evaluation
        dataset: The dataset of examples

    Returns:
        A tuple(avg. loss, avg. accuracy)
    """
    model.eval()
    device = "cuda" if torch.cuda.is_available() else "cpu"
    running_loss = 0.0
    running_accuracy = 0.0
    dataloader = torch.utils.data.DataLoader(dataset)
    with torch.no_grad():
        for batch, expected in iter(dataloader):
            predicted = model(batch.to(device))
            expected = expected.to(device)
            running_loss += criterion(predicted, expected).item()
            running_accuracy += accuracy(expected, predicted)
    return running_loss / len(dataset), running_accuracy / len(dataset)


def eval_checkpoint(
    model: torch.nn.Module,
    tag: str = "acc",
    num_iters: Optional[int] = None,
    device: Optional[str] = None,
    **kwargs,
) -> float:
    """
    Loads the provided model from the checkpoint at ``epoch`` and runs
    ``evaluate``, returning the result. If ``epoch`` is not provided, the
    latest checkpoint is used.
    """
    from slidingpuzzle.algorithms import evaluate as evaluate_
    from slidingpuzzle.nn.train import load_checkpoint
    from slidingpuzzle.nn.heuristics import set_heuristic

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    load_checkpoint(model, tag=tag)
    model.to(device)
    heuristic = set_heuristic(model)
    if num_iters is None:
        return evaluate_(model.h, model.h, weuristic=heuristic, **kwargs)
    else:
        return evaluate_(
            model.h, model.h, weuristic=heuristic, num_iters=num_iters, **kwargs
        )



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for training a new sliding tile puzzle model for computing heuristic value.
"""

from typing import Collection, Optional

import collections
import itertools
import logging
import math
import random

import numpy as np
import tensorboard
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
import tqdm
from tqdm.contrib.logging import tqdm_logging_redirect

import slidingpuzzle.nn.paths as paths
from slidingpuzzle.nn.dataset import load_or_build_dataset
from slidingpuzzle.nn.eval import accuracy, evaluate


log = logging.getLogger(__name__)


def set_seed(seed: int) -> None:
    """
    Set all standard random seeds.

    Args:
        seed (int): The seed to use
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def get_state_dict(model: nn.Module, optimizer: optim.Optimizer, **kwargs) -> dict:
    return {
        "model_state_dict": model.state_dict(),
        "optimizer_state_dict": optimizer.state_dict(),
        "board_height": model.h,
        "board_width": model.w,
        **kwargs,
    }


def load_checkpoint(
    model: nn.Module,
    tag: str = "acc",
    optimizer: Optional[optim.Optimizer] = None,
) -> dict:
    """
    Loads a model and optimizer state from the specified epoch.
    If no epoch provided, latest trained model is used.
    In addition to the model and optimizer state, a checkpoint contains:

    * ``epoch``: The epoch this checkpoint is from
    * ``test_accuracy``: The test accuracy at that time
    * ``board_width/height``: The board size

    Note:
        The board size is extracted from ``model.h`` and ``model.w``, which are
        expected to be present.

    Args:
        model: The module to load weights into
        optimizer: The optimizer state to restore. If ``None``, ignored.
        tag: The checkpoint tag to load. If ``None``, latest checkpo

    Returns:
        The checkpoint ``dict``. If not found, returns ``dict`` with default values
        that can be used at the start of search.
    """
    checkpoint_path = paths.get_checkpoint_path(model.h, model.w, tag)
    try:
        checkpoint = torch.load(str(checkpoint_path))
        model.load_state_dict(checkpoint["model_state_dict"])
        if optimizer is not None:
            optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        return checkpoint
    except FileNotFoundError:
        log.info("No model checkpoint found.")
        return {"epoch": 0, "test_accuracy": float("-inf")}


def save_checkpoint(state: dict, tag: Optional[str] = None) -> None:
    h = state["board_height"]
    w = state["board_width"]
    checkpoint_path = paths.get_checkpoint_path(h, w, tag)
    torch.save(state, str(checkpoint_path))


def launch_tensorboard(dirname: str = paths.TENSORBOARD_DIR) -> str:
    """
    Launches tensorboard in the specified directory.

    Returns:
        The launch URL.
    """
    tensorboard.program.logger.setLevel(logging.ERROR)
    tb = tensorboard.program.TensorBoard()
    tb.configure(argv=[None, "--logdir", dirname])
    url = tb.launch()
    werkzeug = logging.getLogger("werkzeug")
    werkzeug.setLevel(logging.ERROR)
    return url


def linear_regression_beta(data):
    r"""
    Calculates the :math:`\beta` (slope) of the linear regresion of ``data``.
    Used to determine which direction a trend appears to be moving. The formula
    derived below comes from linear regression. We wish to solve for :math:`\beta`
    in the equation

    .. math::
        y = \alpha + \beta x

    where the formula for :math:`\beta` is

    .. math::
        \beta = \frac{\sum^{n}_{i=1}{(x_i - \overline{x})(y_i - \overline{y})}}
            {\sum^{n}_{i=1}{(x_i - \overline{x})^2}}
            \text{.}

    In our case, the values for :math:`x` are the range :math:`[1, n]`, where
    :math:`n` is ``len(data)``. We can compute the sum of this range using the
    standard formula,

    .. math::
        n(n+1)/2

    and to compute the average :math:`\overline{x}`,

    .. math::
        \overline{x} = n(n+1)/2n = (n+1)/2 \text{.}
    """
    x_mean = (len(data) + 1) / 2
    y_mean = sum(data) / len(data)
    numerator = sum((x - x_mean) * (y - y_mean) for x, y in zip(range(len(data)), data))
    denominator = sum((x - x_mean) ** 2 for x in data)
    return numerator / denominator


def train(
    model: nn.Module,
    num_examples: int = 2**14,
    dataset: Optional[torch.utils.data.Dataset] = None,
    train_fraction: float = 0.95,
    num_epochs: int = 0,
    batch_size: int = 256,
    early_quit_epochs: int = 256,
    early_quit_threshold: float = 1e-5,
    device: Optional[str] = None,
    checkpoint_freq: int = 50,
    tensorboard_dir: str = paths.TENSORBOARD_DIR,
    seed: int = 0,
    **kwargs,
) -> None:
    """
    Trains a model for ``num_epochs``. If no prior model is found, a new one is
    created. If a prior model is found, training will be resumed. If no dataset is
    found, one will be built. Will automatically launch tensorboard and checkpoint
    periodically. The train/test split is done randomly, but in a reproducible way
    so that training may be paused and resumed without leaking test data.

    Note:
        If you change your dataset between runs, the train/test split will no longer
        be consistent with original training, so training will need to be restarted
        to prevent leaking test data.

    Args:
        model: The model instance to train
        num_examples: Total number of examples to use for training / test. Ignored
            if ``dataset`` is provided
        dataset: A custom dataset to use. If not provided, it will be generated.
        train_fraction: The train/test split
        num_epochs: Total number of epochs model should be trained for. Use 0 to run
            until the ``early_quit_epochs`` logic is hit.
        batch_size: Batch size to use for training
        early_quit_epochs: We will hold an ``early_quit_epochs``-sized window of test
            loss values. If the linear regression slope of these data points is
            > ``early_quit_threshold``, we will terminate training early. Use 0 to
            disable this feature and always run until ``num_epochs`` of training have
            completed. If both this and ``num_epochs`` are 0, training will run until
            manually interrupted.
        early_quit_beta: If the slope of the test loss rises above this value, training
            is terminated.
        device: Device to train on. Default will autodetect CUDA or use CPU
        checkpoint_freq: Model will be checkpointed each time this many epochs
            elapse. If 0, no epoch checkpointint will be used. (Highest test acc. will
            still be checkpointed and also final checkpoint on termination.)
        tensorboard_dir: The root tensorboard dir for logs. Default is "tensorboard".
        seed: Seed used for torch random utilities for reproducibility
        kwargs: Additional args that may be passed to
            :func:`slidingpuzzle.nn.dataset.make_examples` when generating a new
            dataset. Can be used to customize the search algorithm used to find
            training examples if, for example, it is taking too long.
    """
    set_seed(seed)
    h, w = model.w, model.h

    # prepare tensorboard to record training
    url = launch_tensorboard(tensorboard_dir)
    log.info(f"Tensorboard launched: {url}")
    log_dir = paths.get_log_dir(tensorboard_dir, h, w)
    comment = f"EXAMPLES_{num_examples}_BS_{batch_size}"
    writer = SummaryWriter(log_dir, comment=comment)
    layout = {
        model.version: {
            "loss": ["Multiline", ["loss/train", "loss/test"]],
            "accuracy": ["Multiline", ["accuracy/train", "accuracy/test"]],
        }
    }
    writer.add_custom_scalars(layout)

    # prepare dataset
    train_size = math.floor(num_examples * train_fraction)
    test_size = num_examples - train_size
    if dataset is None:
        dataset = load_or_build_dataset(h, w, num_examples, **kwargs)
    train_dataset, test_dataset = torch.utils.data.random_split(
        dataset, [train_size, test_size], generator=torch.Generator().manual_seed(seed)
    )

    # prepare model
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    optimizer = optim.SGD(model.parameters(), lr=0.0001, momentum=0.9)
    highest_acc = load_checkpoint(model, tag="acc", optimizer=optimizer)[
        "test_accuracy"
    ]
    epoch = load_checkpoint(model, tag="latest", optimizer=optimizer)["epoch"]
    criterion = nn.MSELoss(reduction="sum")
    test_loss_window: Collection[float] = collections.deque(maxlen=early_quit_epochs)

    # prepare training loop
    if 0 == num_epochs:
        log.info(f"initial epoch={epoch}, batch_size={batch_size}")
        epoch_iter = itertools.count(start=epoch)
    else:
        log.info(
            f"initial epoch={epoch}/{num_epochs}, "
            f"batch_size={batch_size}, "
            f"highest_acc={highest_acc}"
        )
        epoch_iter = range(epoch, num_epochs)

    total_epochs = num_epochs if num_epochs else None
    pbar = tqdm.tqdm(total=total_epochs, initial=epoch)
    try:
        for epoch in epoch_iter:
            model.train()  # every epoch b/c eval happens every epoch
            running_loss = 0.0
            running_accuracy = 0.0
            num_train_examples = 0

            for batch, expected in iter(
                DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
            ):
                num_train_examples += batch.shape[0]
                batch = batch.to(device)
                expected = expected.to(device)
                optimizer.zero_grad()
                predicted = model(batch)
                loss = criterion(predicted, expected)
                loss.backward()
                optimizer.step()
                running_loss += loss.item()
                running_accuracy += accuracy(expected, predicted)

            # need to normalize loss and acc for the examples we actually saw
            running_loss /= num_train_examples
            running_accuracy /= num_train_examples
            test_loss, test_accuracy = evaluate(model, criterion, test_dataset)
            test_loss_window.append(test_loss)
            writer.add_scalar("loss/training", running_loss, epoch)
            writer.add_scalar("accuracy/training", running_accuracy, epoch)
            writer.add_scalar("loss/test", test_loss, epoch)
            writer.add_scalar("accuracy/test", test_accuracy, epoch)
            pbar.set_description(f"test/acc: {test_accuracy:0.5f}")
            pbar.update(1)

            if test_accuracy > highest_acc:
                # save a tagged checkpoint for highest acc
                highest_acc = test_accuracy
                state = get_state_dict(
                    model, optimizer, epoch=epoch, test_accuracy=test_accuracy
                )
                save_checkpoint(state, "acc")
            if checkpoint_freq > 0 and epoch % checkpoint_freq == checkpoint_freq - 1:
                # save latest model state + epoch labeled checkpoint
                state = get_state_dict(
                    model, optimizer, epoch=epoch, test_accuracy=test_accuracy
                )
                save_checkpoint(state, f"epoch_{epoch}")
            if early_quit_epochs > 0 and len(test_loss_window) == early_quit_epochs:
                # if we are using early quitting, check the overall trendline is down
                if linear_regression_beta(test_loss_window) > early_quit_threshold:
                    with tqdm_logging_redirect():
                        log.info(
                            f"Early quit threshold reached at epoch {epoch}, "
                            f"highest_acc={highest_acc}"
                        )
                    break
            if num_epochs and epoch == num_epochs - 1:
                with tqdm_logging_redirect():
                    log.info("Max epochs reached.")
                break

    except KeyboardInterrupt:
        with tqdm_logging_redirect():
            log.info("Training interrupted.")
    finally:
        pbar.close()

    # save final state in case we were interrupted
    state = get_state_dict(model, optimizer, epoch=epoch, test_accuracy=test_accuracy)
    save_checkpoint(state, "latest")



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A collection of functions for working with sliding tile puzzle boards.
"""

from typing import Iterable, Iterator, Optional, TypeAlias

import itertools
import random
import sys

import numpy as np
import numpy.typing as npt


BLANK_TILE = 0
Board: TypeAlias = npt.NDArray
FrozenBoard: TypeAlias = tuple[tuple[int, ...], ...]


def new_board(h: int, w: int, dtype: Optional[npt.DTypeLike] = None) -> Board:
    """
    Create a new board in the default solved state.

    Args:
        h: Height of the board.
        w: Width of the board.
        dtype: An optional numpy dtype for the board. Will be inferred if not provided.

    Returns:
        The new board.
    """
    board = np.arange(1, 1 + h * w, dtype=dtype).reshape(h, w)
    board[-1, -1] = BLANK_TILE
    return board


def board_from(*rows: Iterable[int], dtype: Optional[npt.DTypeLike] = None) -> Board:
    r"""
    Constructs a new board from a provided iterable of rows.
    May throw a `ValueError` if rows are unequal in size or duplicate values
    are found.

    Args:
        rows: One or more rows that constitute a board. Each row should contain ints.
        dtype: An optional numpy dtype for the board. Will be inferred if not provided.

    Raises:
        TypeError: If a non-int value is found in a row.
        ValueError: If rows have unequal length or tiles are duplicated, missing, or
            have unexpected gaps.

    Returns:
        The constructed board.
    """
    board = []
    for row in rows:
        new_row = []
        for tile in row:
            # ensure all values are ints
            if not (isinstance(tile, int) or isinstance(tile, np.integer)):
                raise TypeError(
                    f"Expected tile to be an int, but received '{type(tile)}'."
                )
            new_row.append(tile)
        board.append(new_row)

    # sanity check row lengths
    for r1, r2 in zip(board[:-1], board[1:]):
        if len(r1) != len(r2):
            raise ValueError(
                f"All rows must have the same length ({len(r1)} != {len(r2)})."
            )

    # check all values are present
    tiles = list(sorted(itertools.chain(*board)))
    if tiles != list(range(len(board) * len(board[0]))):
        raise ValueError("Your board is missing tiles or contains duplicates.")

    return np.array(board, dtype=dtype)


def board_from_iter(h: int, w: int, iter: Iterable[int]) -> Board:
    r"""
    Given an iterable of ints, will construct a board of size ``h x w`` in
    row-major order. The number of values should equal :math:`h \cdot w` and
    all values must be provided. (Any partial extra row will be discarded.)

    Args:
        h: Height of the board to construct
        w: Width of the board to construct
        iter: Iterable of values to construct board from

    Raises:
        TypeError: If a non-int value is found in a row.
        ValueError: If rows have unequal length or tiles are duplicated, missing, or
            have unexpected gaps.

    Returns:
        The constructed board.
    """
    # construct board from iterable
    rows = []
    row = []
    for tile in iter:
        row.append(tile)
        if len(row) == w:
            rows.append(row)
            row = []
    return board_from(*rows)


def flatten_board(board: Board | FrozenBoard) -> list[int]:
    """
    Flattens a board to a list. Useful for quickly compressing the board
    state. Can be reconstructed using :func:`board_from_iter`.

    Args:
        board: The board to flatten

    Returns:
        A flat sequence of ints representing the board.
    """
    return [tile for row in board for tile in row]


def freeze_board(board: Board) -> FrozenBoard:
    """
    Obtain a frozen copy of the board that is hashable.

    Args:
        board: The board to freeze.

    Returns:
        A frozen copy of the board.
    """
    return tuple(tuple(row) for row in board)


def print_board(board: Board | FrozenBoard, file=sys.stdout) -> Board | FrozenBoard:
    """
    Convienance function for printing a formatted board.

    Args:
        board: The board to print.
        file: The target output file. Defaults to stdout.

    Returns:
        The input board (useful for call chaining).
    """
    board_size = len(board) * len(board[0])
    # the longest str we need to print is the largest tile number
    max_width = len(str(board_size - 1))
    for row in board:
        for tile in row:
            if tile == BLANK_TILE:
                print(" " * max_width, end=" ", file=file)
            else:
                print(str(tile).ljust(max_width), end=" ", file=file)
        print(file=file)

    return board


def get_goal_y(h: int, w: int, tile: int) -> int:
    """
    Given a board width and tile number, returns the goal row position of ``tile``.

    Args:
        h: Height of the board
        w: Width of the board
        tile: The tile number of interest

    Returns:
        The goal row (y-coord)
    """
    if BLANK_TILE == tile:
        return h - 1
    return (tile - 1) // w


def get_goal_x(h: int, w: int, tile: int) -> int:
    """
    Given a board width and tile number, returns the goal column position of ``tile``.

    Args:
        h: Height of the board
        w: Width of the board
        tile: The tile number of interest

    Returns:
        The goal column (x-coord)
    """
    if BLANK_TILE == tile:
        return w - 1
    return (tile - 1) % w


def get_goal_yx(h: int, w: int, tile: int) -> tuple[int, int]:
    """
    Given a board width and tile number, returns the goal (y, x) position of ``tile``.

    Args:
        h: Height of the board
        w: Width of the board
        tile: The tile number of interest

    Returns:
        A tuple (y, x) indicating the tile's goal position.
    """
    return get_goal_y(h, w, tile), get_goal_x(h, w, tile)


def get_goal_tile(h: int, w: int, pos: tuple[int, int]) -> int:
    """
    Given a board width and (y, x)-coord, return the tile number that belongs at (y, x).

    Args:
        h: Height of the board
        w: Width of the board
        pos: A (y, x)-position on the board.

    Returns:
        The tile number that belongs at (y, x).
    """
    if pos == (h - 1, w - 1):
        return BLANK_TILE
    return (w * pos[0]) + pos[1] + 1


def is_solved(board: Board) -> bool:
    """
    Determine if a board is in the solved state.

    Args:
        board: The board

    Returns:
        True if the board is solved, otherwise False.
    """
    h, w = board.shape
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            if tile != get_goal_tile(h, w, (y, x)):
                return False
    return True


def find_tile(board: Board | FrozenBoard, tile: int) -> tuple[int, int]:
    """
    Given a tile number, find the (y, x)-coord on the board.

    Args:
        board: The puzzle board.
        move: The

    Raises:
        ValueError: If the tile is not on the board.

    Returns:
        The tile position.
    """
    if isinstance(board, np.ndarray):
        y, x = np.where(board == tile)
        return y[0], x[0]
    for y, row in enumerate(board):
        for x in range(len(row)):
            if board[y][x] == tile:
                return y, x
    raise ValueError(f'There is no tile "{tile}" on the board.')


def find_blank(board: Board | FrozenBoard) -> tuple[int, int]:
    """
    Locate the blank tile's (y, x)-coord.
    Equivalent to ``find_tile(board, BLANK_TILE)``.

    Args:
        board: The puzzle board.

    Returns:
        The (y, x)-coord of the blank tile.
    """
    return find_tile(board, BLANK_TILE)


def get_next_moves(
    board: Board | FrozenBoard,
    blank_pos: Optional[tuple[int, int]] = None,
) -> list[tuple[int, int]]:
    """
    Return a list of all possible moves.

    Args:
        board: The current puzzle board.
        blank_pos: The position of the empty tile.
            If not provided, it will be located automatically.

    Returns:
        A list of (y, x)-coords that are tile positions capable of
        swapping with the blank tile.
    """
    if blank_pos is None:
        blank_pos = find_blank(board)
    y, x = blank_pos
    moves = []
    for dy, dx in ((0, -1), (0, 1), (-1, 0), (1, 0)):
        if 0 <= y + dy < len(board) and 0 <= x + dx < len(board[0]):
            moves.append((y + dy, x + dx))
    return moves


def swap_tiles(
    board: Board,
    tile1: tuple[int, int] | int,
    tile2: Optional[tuple[int, int] | int] = None,
) -> Board:
    """
    Mutates the board by swapping a pair of tiles.

    Args:
        board: The board to modify.
        tile1: The first tile position or value.
        tile2: The second tile position or value. If None, the blank will be located
            and used.

    Return:
        The modified board, for conveience chaining calls.
    """
    if isinstance(tile1, int) or isinstance(tile1, np.integer):
        tile1 = find_tile(board, tile1)
    if isinstance(tile2, int) or isinstance(tile2, np.integer):
        tile2 = find_tile(board, tile2)
    elif tile2 is None:
        tile2 = find_blank(board)

    y1, x1 = tile1
    y2, x2 = tile2
    board[y1, x1], board[y2, x2] = board[y2, x2], board[y1, x1]

    return board


def random_move(board: Board, blank_pos: Optional[tuple[int, int]] = None) -> Board:
    """
    Picks a random legal move and applies it to the board.

    Args:
        board: The board
        blank_pos: The position of the blank. If not provided, will be found.
    """
    moves = get_next_moves(board, blank_pos)
    return swap_tiles(board, random.choice(moves), blank_pos)


def count_inversions(board: Board) -> int:
    """
    From each tile, count the number of tiles that are out of place after this tile.
    Returns the sum of all counts. See :func:`is_solvable`.

    Args:
        board: The puzzle board.

    Returns:
        The count of inversions.
    """
    _, w = board.shape
    board_size = np.prod(board.shape)
    inversions = 0
    for i in range(board_size):
        t1 = board[i // w, i % w]
        if t1 == BLANK_TILE:
            continue
        for j in range(i + 1, board_size):
            t2 = board[j // w, j % w]
            if t2 == BLANK_TILE:
                continue
            if t2 < t1:
                inversions += 1
    return inversions


def is_solvable(board: Board) -> bool:
    """
    Determines if it is possible to solve this board.

    Note:
        The algorithm counts `inversions`_ to determine solvability.
        The "standard" algorithm has been modified here to support
        non-square board sizes.

    Args:
        board: The puzzle board.

    Returns:
        bool: True if the board is solvable, False otherwise.

    .. _inversions:
        https://www.cs.princeton.edu/courses/archive/spring21/cos226/assignments/8puzzle/specification.php
    """
    inversions = count_inversions(board)
    h, w = board.shape
    if w % 2 == 0:
        y, _ = find_blank(board)
        if h % 2 == 0:
            if (inversions + y) % 2 != 0:
                return True
        else:
            if (inversions + y) % 2 == 0:
                return True
    elif inversions % 2 == 0:
        return True
    return False


def shuffle_board(board: Board) -> Board:
    """
    Shuffles a board (in place). Board is always solvable.

    Args:
        board: The board to shuffle.

    Returns:
        The shuffled board.
    """
    h, w = board.shape
    while True:
        for y in range(h):
            for x in range(w):
                pos1 = y, x
                pos2 = random.randrange(h), random.randrange(w)
                swap_tiles(board, pos1, pos2)
        if is_solvable(board):
            return board


def shuffle_board_lazy(
    board: Board, num_moves: Optional[int] = None, moves: Optional[list] = None
) -> Board:
    """
    Shuffles a board in place by making random legal moves.
    Each move is first checked to avoid repeated states, although
    this does not guarantee the

    Args:
        board: The board to shuffle.
        num_moves (int): Number of random moves to make.
            If ``None``, ``(h + w) * 2`` will be used.
        moves: If a list is provided, the moves made will be appended.

    Returns:
        The same board for chaining convience.
    """
    h, w = board.shape
    if num_moves is None:
        num_moves = (h + w) * 2
    blank_pos = find_blank(board)
    visited: set[FrozenBoard] = set()
    for _ in range(num_moves):
        next_moves = get_next_moves(board, blank_pos)
        random.shuffle(next_moves)
        new_move = False  # tracks if this is a new move
        # try to find a new move
        while next_moves:
            next_move = next_moves.pop()
            swap_tiles(board, blank_pos, next_move)
            # have we been here before?
            if visit(visited, board):
                swap_tiles(board, blank_pos, next_move)  # undo move
            else:
                new_move = True
                break

        # if we couldn't find a new move, we're done
        if not new_move:
            return board

        # if we're tracking which moves we made, record this move
        if moves is not None:
            moves.append(next_move)

        # update blank_pos
        blank_pos = next_move
    return board


def solution_as_tiles(board: Board, solution: list[tuple[int, int]]) -> list[int]:
    """
    Converts a list of (y, x)-coords indicating moves into tile numbers,
    given a starting board configuration.

    Args:
        board: The initial board we will apply moves to (does not alter board).
        solution: A list of move coordinates in (y, x) form.

    Returns:
        A list of ints, which indicate a sequence of tile numbers to move.
    """
    board = np.copy(board)
    tiles = []
    blank_pos = find_blank(board)
    for move in solution:
        y, x = move
        tiles.append(board[y][x])
        swap_tiles(board, blank_pos, move)
        blank_pos = move
    return tiles


def visit(visited: set[FrozenBoard], board: Board) -> bool:
    """
    Helper to check if this state already exists. Otherwise, record it.
    Returns True if we have already been here, False otherwise.

    Args:
        visited: Set of boards already seen.
        board: The current board.

    Returns:
        True if we've been here before.
    """
    frozen_board = freeze_board(board)
    if frozen_board in visited:
        return True
    visited.add(frozen_board)
    return False


def board_generator(h: int, w: int) -> Iterator[Board]:
    """
    Returns a generator that yields all solvable boards in lexicographical order.

    Args:
        h: Height of board
        w: Width of board

    Yields:
        A board permutation
    """
    for values in itertools.permutations(range(h * w)):
        board = board_from_iter(h, w, values)
        if is_solvable(board):
            yield board



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides heuristic functions used to evaluate board states.

Each function accepts a puzzle board as input, and returns a numeric value
that indicates the estimated distance from the goal.
"""

from typing import Callable, Iterator, TypeAlias

import logging
import math

import numpy as np
import numpy.typing as npt

from slidingpuzzle.board import (
    BLANK_TILE,
    Board,
    find_blank,
    find_tile,
    get_goal_tile,
    get_goal_y,
    get_goal_x,
    get_goal_yx,
    freeze_board,
    is_solved,
    swap_tiles,
)


Heuristic: TypeAlias = Callable[[Board], int | float]

log = logging.getLogger(__name__)
manhattan_tables = {}  # (h, w): {(r, c, tile): manhattan dist}


def euclidean_distance(board: Board) -> float:
    r"""
    The distance between each tile and its destination, as measured in Euclidean space

    .. math::
        \sum_{i}^{n} \sqrt{(\text{tile}_{i, x} - \text{goal}_{i, x})^2 +
            (\text{tile}_{i, y} - \text{goal}_{i, y})^2}

    Args:
        board: The board

    Returns:
        The sum of all tile distances from their goal positions
    """
    h, w = board.shape
    dist = 0.0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            if BLANK_TILE == tile:
                continue
            goal_y, goal_x = get_goal_yx(h, w, tile)
            a, b = abs(y - goal_y), abs(x - goal_x)
            dist += math.sqrt(a**2 + b**2)
    return dist


def hamming_distance(board: Board) -> int:
    r"""
    The count of misplaced tiles.

    .. math::
        \sum_{i}^{n} \text{tile}_i \oplus \text{goal}_i

    Args:
        board: The board

    Returns:
        The number of tiles out of their goal positions
    """
    h, w = board.shape
    dist = 0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            if BLANK_TILE == tile:
                continue
            goal_y, goal_x = get_goal_yx(h, w, tile)
            if y != goal_y or x != goal_x:
                dist += 1
    return dist


def last_moves_distance(board: Board) -> int:
    r"""
    If the tiles required for the final move are not in appropriate positions, we must
    make at least 2 more moves to shuffle them. The constraints are weakened so that
    this heuristic can be constructively combined with :func:`manhattan_distance`
    without violating admissibility, as these are more powerful together than alone.

    Args:
        board: The board

    Returns:
        2 if the both of the last two tiles adj. to the blank corner are outside of the
        blank row/col.
    """
    h, w = board.shape
    dist = 0
    adj1 = get_goal_tile(h, w, (h - 2, w - 1))  # tile in row next to goal
    adj2 = get_goal_tile(h, w, (h - 1, w - 2))  # tile in col next to goal
    adj1_y, _ = find_tile(board, adj1)  # coords of these tiles on cur. board
    _, adj2_x = find_tile(board, adj2)
    # if the tiles that must be the final move are not located in the goal row/col,
    # there will need to be additional moves made to shuffle them before the final move
    if adj1_y != h - 1 and adj2_x != w - 1 and BLANK_TILE != board[-1, -1]:
        dist += 2
    return dist


def corner_tiles_distance(board: Board, ignore_conflicts: bool = False) -> int:
    r"""
    If tiles that belong in corners are not there, but the adjacent tiles are correct,
    additional reshuffling will need to occur. We must ensure that row/col conflicts
    have not already accounted for this in order to combine admissibily with the
    :func:`linear_conflict_distance`.

    Args:
        board: The board
        ignore_conflicts: If True, tiles adjacent to corners that have conflicts in
            their row/col will still be counted. When True, this cannot be admissibly
            combined with :func:`linear_conflict_distance`.
    """
    h, w = board.shape
    dist = 0

    def has_row_conflict(y, x, skip_x) -> bool:
        """
        On row `y`, we check every tile for conflict with column `x` (ignoring
        `skip_x`). The skip variable is used to ignore the corner column.
        We need to know if column `x` has any conflicts, b/c if it does, we cannot
        add additional distance
        """
        for col1 in range(w):
            tile1 = board[y, col1]
            if BLANK_TILE == tile1 or col1 == skip_x:
                continue
            for col2 in range(col1 + 1, w):
                tile2 = board[y, col2]
                # skip tiles that don't involve our x of interest
                if BLANK_TILE == tile2 or col2 == skip_x or (col1 != x and col2 != x):
                    continue
                if tile2 < tile1:
                    return True
        return False

    def has_col_conflict(y, x, skip_y) -> bool:
        for row1 in range(h):
            tile1 = board[row1, x]
            if BLANK_TILE == tile1 or row1 == skip_y:
                continue
            for row2 in range(row1 + 1, h):
                tile2 = board[row2, x]
                if BLANK_TILE == tile2 or row2 == skip_y or (row1 != y and row2 != y):
                    continue
                if tile2 < tile1:
                    return True
        return False

    def has_conflict(y, x, corner) -> bool:
        if ignore_conflicts:
            return False
        corner_y, corner_x = corner
        return has_row_conflict(y, x, corner_x) or has_col_conflict(y, x, corner_y)

    def check_adjacent(y, x, corner) -> int:
        if get_goal_tile(h, w, (y, x)) == board[y, x] and not has_conflict(
            y, x, corner
        ):
            return 2
        return 0

    def check_corner(y, x, adj1, adj2) -> int:
        if board[y, x] != BLANK_TILE and board[y, x] != get_goal_tile(h, w, (y, x)):
            return check_adjacent(*adj1, (y, x)) + check_adjacent(*adj2, (y, x))
        return 0

    # top-left corner
    dist += check_corner(0, 0, (0, 1), (1, 0))
    # top-right corner
    dist += check_corner(0, w - 1, (0, w - 2), (1, w - 1))
    # bottom-left corner
    dist += check_corner(h - 1, 0, (h - 1, 1), (h - 2, 0))

    return dist


def linear_conflict_distance(board: Board) -> int:
    r"""
    Starts with Manhattan distance, then for each row and column, the number of tiles
    "in conflict" are identified, and 2 * this number is added to the distance.
    (It will take at least 2 additional moves to reshuffle the conflicting tiles into
    their correct positions.) This is an admissible improvement over
    :func:`manhattan_distance` (`Hansson, Mayer, Young, 1985 <hannson_>`_).
    The :func:`last_moves_distance` and :func:`corner_tiles_distance` are included
    (`Korf and Taylor, 1996 <korf_>`_).

    Args:
        board: The board

    Returns:
        Estimated distance to goal.

    .. _hannson:
        https://academiccommons.columbia.edu/doi/10.7916/D8154QZT/download

    .. _korf:
        https://www.aaai.org/Library/AAAI/1996/aaai96-178.php
    """
    h, w = board.shape
    dist = manhattan_distance(board)

    def line_generator() -> Iterator[tuple[npt.NDArray, npt.NDArray]]:
        r"""
        Helper to generate all lines on a board (rows followed by columns) along with
        a bool specifying whether the tile is in its goal line.

        Returns:
            A tuple of line index, line values, and goal positions..
        """
        for y, row in enumerate(board):
            yield row, np.fromiter(
                (y == get_goal_y(h, w, tile) for tile in row), dtype=bool
            )
        for x, col in enumerate(board.T):
            yield col, np.fromiter(
                (x == get_goal_x(h, w, tile) for tile in col), dtype=bool
            )

    def get_line_conflicts(line, goals) -> npt.NDArray[np.integer]:
        r"""
        Helper to return a list of conflict counts for a line.

        Args:
            line: The values in the line
            goals: The goal positions for each tile in line

        Returns:
            Conflict counts for each tile
        """
        conflicts = np.zeros(len(line), dtype=int)
        for pos1, (tile1, goal1) in enumerate(zip(line, goals)):
            # check if this tile is in it's goal line
            if not goal1 or BLANK_TILE == tile1:
                continue
            # now check tiles after this
            next_pos = pos1 + 1
            for pos2, (tile2, goal2) in enumerate(
                zip(line[next_pos:], goals[next_pos:]), next_pos
            ):
                # same checks for tile2
                if not goal2 or BLANK_TILE == tile2:
                    continue
                # check if these tiles are in conflict
                if tile2 < tile1:
                    conflicts[pos1] += 1
                    conflicts[pos2] += 1
        return conflicts

    for line, goals in line_generator():
        line = np.copy(line)  # don't modify original board
        while np.any(line_conflicts := get_line_conflicts(line, goals)):
            dist += 2
            line[np.argmax(line_conflicts)] = BLANK_TILE

    dist += last_moves_distance(board)
    dist += corner_tiles_distance(board)

    return dist


def prepare_manhattan_table(h, w) -> dict[tuple[int, int, int], int]:
    log.debug(f"Preparing first use of Manhattan table {h}x{w}x{h*w}...")
    table = {}
    for y in range(h):
        for x in range(w):
            for tile in range(h * w):
                if BLANK_TILE == tile:
                    table[(y, x, tile)] = 0
                else:
                    goal_y, goal_x = get_goal_yx(h, w, tile)
                    table[(y, x, tile)] = abs(y - goal_y) + abs(x - goal_x)
    return table


def manhattan_distance(board: Board) -> int:
    r"""
    The minimum number of moves needed to restore the board to the goal state, if tiles
    could be moved through each other.

    .. math::
        \sum_{i}^{n} |\text{tile}_{i, x} - \text{goal}_{i, x}| +
            |\text{tile}_{i, y} - \text{goal}_{i, y}|

    Args:
        board: The board

    Returns:
        The sum of all tile distances from their goal positions
    """
    h, w = board.shape
    table = manhattan_tables.get((h, w), None)
    if table is None:
        table = prepare_manhattan_table(h, w)
        manhattan_tables[(h, w)] = table

    dist = 0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            dist += table[(y, x, tile)]
    return dist


def random_distance(board: Board) -> int:
    r"""
    A random distance computed as a hash of the board state. Useful as a baseline.

    .. math::
        hash(\text{board})

    Args:
        board: The board

    Returns:
        A random distance that is consistent for a given board state
    """
    return hash(freeze_board(board))


def relaxed_adjacency_distance(board: Board) -> int:
    r"""
    Repeatedly swap misplaced tiles into their goal positions, ignoring legal board
    distance and rules. We repeat this process until the board is solved. This
    heuristic is a slight improvement to Hamming distance.

    Args:
        board: The board

    Returns:
        The estimated distance to goal.
    """
    h, w = board.shape
    board = np.copy(board)
    dist = 0

    def swap_first_misplaced(blank_yx: tuple[int, int]) -> None:
        for y in range(h):
            for x in range(w):
                if get_goal_yx(h, w, board[y][x]) != (y, x):
                    swap_tiles(board, (y, x), blank_yx)

    while not is_solved(board):
        blank_yx = find_blank(board)
        if blank_yx == get_goal_yx(h, w, BLANK_TILE):
            swap_first_misplaced(blank_yx)
        else:
            swap_tiles(board, get_goal_tile(h, w, blank_yx), blank_yx)
        dist += 1

    return dist



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines neural network-guided heuristics.
"""
import torch

import slidingpuzzle.nn.models as models
import slidingpuzzle.nn.paths as paths
from slidingpuzzle.board import Board
from slidingpuzzle.heuristics import Heuristic


MODEL_HEURISTICS = {}


def get_heuristic_key(h: int, w: int, version: str):
    board_size_str = paths.get_board_size_str(h, w)
    return f"{board_size_str}_{version}"


def make_heuristic(model: torch.nn.Module | torch.ScriptModule):
    device = "cuda" if torch.cuda.is_available() else "cpu"

    def heuristic(board: tuple[list[int], ...]) -> float:
        tensor = torch.tensor(board, dtype=torch.float32).unsqueeze(0).to(device)
        return model(tensor).item()

    return heuristic


def set_heuristic(model: torch.nn.Module | torch.ScriptModule):
    key = get_heuristic_key(model.w, model.h, model.version)
    heuristic = make_heuristic(model)
    MODEL_HEURISTICS[key] = heuristic
    return heuristic


def get_heuristic(h: int, w: int, version: str) -> Heuristic:
    key = get_heuristic_key(h, w, version)
    heuristic = MODEL_HEURISTICS.get(key, None)
    if heuristic is None:
        model = models.load_model(h, w, version)
        key = get_heuristic_key(h, w, version)
        heuristic = make_heuristic(model)
        MODEL_HEURISTICS[key] = heuristic
    return heuristic


##################################################################
# methods beyond here correspond to predefined model classes
# you can add your own model heuristics below
##################################################################


def v1_distance(board: Board) -> float:
    """
    A neural network that estimates distance to goal

    Args:
        board: The board

    Returns:
        An estimated number of moves to reach the goal
    """
    h, w = len(board), len(board[0])
    heuristic = get_heuristic(h, w, models.VERSION_1)
    return heuristic(board)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for creating, saving, and loading board datasets.
"""

from typing import Optional, TypeAlias

import json
import logging

import torch
import torch.utils
import tqdm

import slidingpuzzle.algorithms as algorithms
import slidingpuzzle.board as board_
import slidingpuzzle.nn.paths as paths
from slidingpuzzle.board import FrozenBoard


Example: TypeAlias = tuple[FrozenBoard, int | float]
log = logging.getLogger(__name__)


class SlidingPuzzleDataset(torch.utils.data.Dataset):
    def __init__(self, examples: list[Example]) -> None:
        super().__init__()
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, idx):
        x, y = self.examples[idx]
        return (
            torch.tensor(x, dtype=torch.float32),
            torch.tensor([y], dtype=torch.float32),
        )


def make_examples(
    h: int,
    w: int,
    num_examples: int,
    ignore_examples: Optional[list[Example]] = None,
    **kwargs,
) -> list[Example]:
    """
    Constructs a list of training examples, which are tuples of:
        (board, num_moves_to_goal)

    Args:
        h: Height of board.
        w: Width of board.
        num_examples: Number of examples to produce.
        ignore_examples: If any example produced matches one in ``ignore_examples``,
            it will be discarded.
        kwargs: Args to pass to the search algorithm used to find board solutions.

    Returns:
        The list of training examples.
    """
    visited: set[FrozenBoard] = set()
    examples = []
    if ignore_examples is not None:
        dupe_found = False
        for board, _ in ignore_examples:
            if board_.visit(visited, board):
                dupe_found = True
        if dupe_found:
            log.warning("Duplicate found in prior examples.")

    # TODO: parallelize
    with tqdm.tqdm(total=num_examples) as pbar:
        while len(examples) < num_examples:
            board = board_.new_board(h, w)
            board_.shuffle_board(board)
            if board_.visit(visited, board):
                continue

            # find a path to use as an accurate training reference
            result = algorithms.search(board, **kwargs)

            # we can use all intermediate boards as examples
            while len(examples) < num_examples:
                distance = len(result.solution)
                examples.append((board_.freeze_board(board), distance))
                pbar.update(1)
                if not len(result.solution):
                    break
                move = result.solution.pop(0)
                board_.swap_tiles(board, move)
                if board_.visit(visited, board):
                    break

    return examples


def load_examples(h: int, w: int, examples_file: Optional[str] = None) -> list[Example]:
    """
    Loads examples from a JSON file.
    """
    if examples_file is None:
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "rt") as fp:
        return json.load(fp)


def save_examples(
    h: int, w: int, examples: list, examples_file: Optional[str] = None
) -> None:
    """
    Save a list of examples to disk as JSON.
    """
    if examples_file is None:
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "wt") as fp:
        json.dump(examples, fp)


def get_examples(
    h: int, w: int, num_examples: int, prior_examples: list[Example], **kwargs
) -> list[Example]:
    """
    Returns ``num_examples`` total unique examples, starting with ``prior_examples`` if
    they are provided. May truncate or produce new examples as necessary.
    """
    examples = list(prior_examples)
    if len(examples) > num_examples:
        examples = examples[:num_examples]
    if len(examples) < num_examples:
        num_new_examples = num_examples - len(prior_examples)
        log.info(f"Constructing {num_new_examples} new examples...")
        examples.extend(make_examples(h, w, num_new_examples, prior_examples, **kwargs))
    return examples


def load_or_build_dataset(
    h: int,
    w: int,
    num_examples: int,
    examples_file: Optional[str] = None,
    **kwargs,
) -> torch.utils.data.Dataset:
    """
    Loads examples, constructs a SlidingPuzzleDataset from them, and returns it.
    If there is a mismatch between the requested num_examples and the loaded
    examples file, examples may be truncated or new examples may be constructed and
    saved to disk. No duplicate examples are produced.

    Args:
        h: The height of the board to locate a dataset for
        w: The width of the board to locate a dataset for
        num_examples: The total number of examples desired. If there are too many,
            examples will be truncated. If there are too few, new examples will be
            constructed.
        examples_file: The name of the examples file to save or load.
        kwargs: Will be forwaded to :func:`make_examples` if it is called.

    Returns:
        A dataset for the requested puzzle size.
    """
    log.info("Loading dataset...")
    try:
        examples = load_examples(h, w, examples_file)
        log.info(f"Dataset loaded with {len(examples)} examples.")
    except FileNotFoundError:
        log.info("No dataset found.")
        examples = []

    # if we were asked for a different number of examples than we have on hand
    if len(examples) != num_examples:
        new_examples = get_examples(h, w, num_examples, examples, **kwargs)
        # if we made new examples, save them for next time
        if len(new_examples) > len(examples):
            save_examples(h, w, new_examples, examples_file)
            log.info("Dataset saved.")
        examples = new_examples

    return SlidingPuzzleDataset(examples)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from slidingpuzzle import *
from slidingpuzzle.nn import *


def test_make_examples():
    examples = make_examples(3, 3, 100)
    assert len(examples) == 100


def test_make_examples2():
    train_examples = make_examples(3, 3, 1)
    test_examples = make_examples(3, 3, 1, train_examples)
    assert len(train_examples) + len(test_examples) == len(
        set(train_examples) | set(test_examples)
    )



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from slidingpuzzle import *


def test_corner_tiles_distance():
    board = new_board(3, 3)
    assert corner_tiles_distance(board) == 0

    # top-left corner
    board = board_from([8, 5, 3], [4, 2, 6], [7, 1, 0])
    assert corner_tiles_distance(board) == 2

    board = board_from([8, 2, 3], [4, 5, 6], [7, 0, 1])
    assert corner_tiles_distance(board) == 4

    # top-right corner
    board = board_from([1, 2, 8], [4, 6, 5], [7, 0, 3])
    assert corner_tiles_distance(board) == 2

    board = board_from([1, 2, 8], [3, 5, 6], [4, 0, 7])
    assert corner_tiles_distance(board) == 4

    # bottom-left corner
    board = board_from([1, 2, 3], [4, 7, 6], [5, 0, 8])
    assert corner_tiles_distance(board) == 2

    board = board_from([1, 2, 3], [4, 7, 6], [5, 8, 0])
    assert corner_tiles_distance(board) == 4


def test_euclidean_distance():
    board = new_board(3, 5)
    assert euclidean_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    c = math.sqrt((2 * 1) ** 2 + (2 * 2) ** 2)
    assert euclidean_distance(board) == c


def test_hamming_distance():
    board = new_board(5, 3)
    assert hamming_distance(board) == 0

    swap_tiles(board, (0, 1), (0, 0))
    assert hamming_distance(board) == 2


def test_last_moves_distance():
    board = board_from([1, 2, 3], [4, 5, 6], [7, 0, 8])
    assert last_moves_distance(board) == 0

    board = board_from([8, 2, 3], [4, 5, 6], [7, 0, 1])
    assert last_moves_distance(board) == 2


def test_linear_conflict_distance():
    board = board_from([1, 2, 3], [4, 5, 6], [7, 0, 8])
    assert linear_conflict_distance(board) == 1

    board = board_from([2, 1, 3], [4, 5, 6], [7, 8, 0])
    assert linear_conflict_distance(board) == 6

    board = board_from([4, 2, 3], [1, 5, 6], [7, 8, 0])
    assert linear_conflict_distance(board) == 6

    board = board_from([2, 1, 3], [4, 5, 6], [7, 0, 8])
    assert linear_conflict_distance(board) == 7

    board = board_from([1, 2, 3], [6, 5, 4], [7, 8, 0])
    assert linear_conflict_distance(board) == 8


def test_manhattan_distance():
    board = new_board(3, 5)
    assert manhattan_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    assert manhattan_distance(board) == 6


def test_relaxed_adjacency_distance():
    h, w = 3, 3
    b = new_board(h, w)
    swap_tiles(b, (h - 1, w - 1), (0, 0))
    assert relaxed_adjacency_distance(b) == 1

    b = new_board(h, w)
    swap_tiles(b, (0, 0), (0, 1))
    swap_tiles(b, (0, 0), (h - 1, w - 1))
    swap_tiles(b, (0, 0), (1, 0))
    assert relaxed_adjacency_distance(b) == 3



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides some convience classes to track search state and results.
"""

from typing import Collection, Optional

import dataclasses

import slidingpuzzle
from slidingpuzzle.board import Board, FrozenBoard


@dataclasses.dataclass(order=True)
class State:
    """
    A dataclass for representing the state during search.
    Although the ``board`` is enough to fully capture the game state,
    it is convenient to track some additional information.

    Args:
        board: The board state.
        blank_pos: The (y, x)-coord of the blank tile.
        history: A list of (y, x)-coords representing moves from the initial
            state to the current board state.
        f: The stored value of this node. Used by some search algorithms.
            The value stored here has no meaning if not using a relevant
            search algorithm.
    """

    board: Board = dataclasses.field(compare=False)
    blank_pos: tuple[int, int] = dataclasses.field(compare=False)
    history: list[tuple[int, int]] = dataclasses.field(
        compare=False, default_factory=list
    )
    f: int | float = 0
    g: int | float = 0  # stored separately for tie-breaking


@dataclasses.dataclass
class SearchResult:
    """
    A dataclass for returning a puzzle solution along with some information
    concerning how the search progressed. Useful for evaluating different
    heuristics.

    Args:
        board: The original input board.
        generated: The number of states generated during search.
        expanded: The number of states evaluated during search.
        unvisited: The list of states that were never reached.
        visited: The set of boards evaluated.
        solution: The list of moves from initial position to solution.

    Note:
        The ``unvisited`` and ``visited`` attributes are traditionally known
        as "open" and "closed" in search parlance. However, Python's built-in
        function ``open`` would collide, so they have been renamed.
    """

    board: Board
    generated: int
    expanded: int
    unvisited: Collection[State]
    visited: set[FrozenBoard]
    solution: Optional[list[tuple[int, int]]]

    def __repr__(self) -> str:
        solution = (
            slidingpuzzle.solution_as_tiles(self.board, self.solution)
            if self.solution
            else "N/A"
        )
        return (
            f"solution={solution}\n"
            f"solution_len={len(self.solution) if self.solution else 'N/A'}, "
            f"generated={self.generated}, "
            f"expanded={self.expanded}, "
            f"unvisited={len(self.unvisited)}, "
            f"visited={len(self.visited)}"
        )

    def __str__(self) -> str:
        return repr(self)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from slidingpuzzle import *


def test_a_star():
    board = new_board(3, 3)
    shuffle_board_lazy(board, num_moves=10)
    assert len(search(board).solution) == len(
        search(board, A_STAR, heuristic=manhattan_distance).solution
    )


@pytest.mark.parametrize("alg", [IDA_STAR, IDDFS])
def test_search_iterative_deepening(alg):
    b = new_board(3, 3)
    shuffle_board_lazy(b, 10)
    # best solution should be found (using default args)
    expected_len = len(search(b).solution)
    actual = search(b, alg=alg, heuristic=manhattan_distance)
    actual_len = len(actual.solution)
    assert expected_len == actual_len


@pytest.mark.parametrize("algorithm", [A_STAR, BEAM, BFS, DFS, GREEDY])
@pytest.mark.parametrize(
    "heuristic",
    [
        euclidean_distance,
        linear_conflict_distance,
        manhattan_distance,
        relaxed_adjacency_distance,
    ],
)
def test_search(algorithm, heuristic):
    board = board_from([5, 2, 4], [3, 0, 1])
    weight = 1
    result = search(
        board,
        alg=algorithm,
        heuristic=heuristic,
        weight=weight,
        width=128,
    )
    assert result.solution is not None
    if algorithm in (A_STAR, BEAM, BFS, IDA_STAR):
        assert len(result.solution) == 15
    print(repr(result))
    print(str(result))


def test_search_hard():
    board = board_from([8, 6, 7], [3, 5, 1], [2, 0, 4])
    result = search(
        board,
        alg=A_STAR,
        heuristic=manhattan_distance,
        weight=1,
    )
    assert len(result.solution) == 27


def test_evaluate():
    assert evaluate(3, 3, num_iters=1) > 0


def test_solution_as_tiles():
    h, w = 3, 3
    b = new_board(h, w)
    swap_tiles(b, (h - 1, w - 1), (h - 2, w - 1))
    swap_tiles(b, (h - 2, w - 1), (h - 2, w - 2))
    r = search(b)
    assert [5, 6] == solution_as_tiles(b, r.solution)


@pytest.mark.slow
def test_heuristic_behavior():
    # we compute avg generated nodes over multiple runs to confirm that
    # heuristic behavior is in line with expectations
    generated_avg = compare(
        3, 3, linear_conflict_distance, manhattan_distance, num_iters=4
    )
    assert generated_avg[0] < generated_avg[1]

    generated_avg = compare(3, 3, manhattan_distance, hamming_distance, num_iters=4)
    assert generated_avg[0] < generated_avg[1]

    # lcd/manhattan/euclidean are good contenders, so we don't compare them
    generated_avg = compare(3, 3, euclidean_distance, hamming_distance, num_iters=4)
    assert generated_avg[0] < generated_avg[1]


@pytest.mark.slow
def test_heuristic_admissibility():
    # validate that solutions are in line with BFS
    # this does not guarantee admissibility, it's just an empirical sanity check
    boards = [shuffle_board(new_board(3, 3)) for _ in range(50)]
    optimal = [len(search(b, "bfs").solution) for b in boards]
    for h in (linear_conflict_distance, manhattan_distance):
        for b, o in zip(boards, optimal):
            assert len(search(b, heuristic=h).solution) == o


@pytest.mark.skip
def test_linear_conflict_distance_exhaustive():
    # test all solvable boards
    for b in board_generator(3, 3):
        expected = len(search(b, heuristic=manhattan_distance).solution)
        actual = len(search(b, heuristic=linear_conflict_distance).solution)
        assert expected == actual, b



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Search algorithms
"""

from typing import Optional

import collections
import heapq

import numpy as np

from slidingpuzzle.board import (
    Board,
    FrozenBoard,
    find_blank,
    get_next_moves,
    is_solvable,
    shuffle_board,
    new_board,
    swap_tiles,
    visit,
)
from slidingpuzzle.heuristics import linear_conflict_distance
from slidingpuzzle.state import State, SearchResult


# all supported algorithms

A_STAR = "a*"
BEAM = "beam"
BFS = "bfs"
DFS = "dfs"
GREEDY = "greedy"
IDA_STAR = "ida*"
IDDFS = "iddfs"
ALGORITHMS = (
    A_STAR,
    BEAM,
    BFS,
    DFS,
    GREEDY,
    IDA_STAR,
    IDDFS,
)


def get_next_states(state: State) -> list[State]:
    """
    Creates a list of next viable states, given the current state.
    """
    moves = get_next_moves(state.board, state.blank_pos)
    next_states = []
    for move in moves:
        # construct altered board
        next_board = np.copy(state.board)
        swap_tiles(next_board, state.blank_pos, move)
        # update history
        next_history = state.history.copy()
        next_history.append(move)
        # after moving, the move is now the blank_pos
        next_state = State(next_board, move, next_history)
        next_states.append(next_state)
    return next_states


def a_star(board: Board, **kwargs) -> SearchResult:
    r"""
    A* heuristic search algorithm.
    Supports weights, depth bounds, and f-bounds.
    The distance of a search state from the goal is computed as:

    .. math::
        f(n) = g(n) + w \cdot h(n)

    or equivalently ``f(n) = len(state.history) + weight * heuristic(state.board)``
    Here, ``g(n)`` is the cost of the solution so far, ``w`` is the weight, and
    ``h(n)`` is the heuristic evaluation. When no heuristic is used, A* becomes
    breadth-first search. When ``weight != 1`` or an inadmissable heuristic is used,
    A* may return a suboptimal solution.

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`.
        f_bound (float): A limit on state cost. Default is :math:`\infty`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.
        heuristic: A function that maps boards to an estimated cost-to-go.
            Default is :func:`slidingpuzzle.heuristics.linear_conflict_distance`.
        weight (float): A constant multiplier on heuristic evaluation

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    f_bound = kwargs.get("f_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)
    heuristic = kwargs.get("heuristic", linear_conflict_distance)
    weight = kwargs.get("weight", 1)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        state = heapq.heappop(unvisited)
        expanded += 1

        # goal check
        if np.array_equal(goal, state.board):
            return SearchResult(
                board, generated, expanded, unvisited, visited, state.history
            )

        # bound
        if len(state.history) > depth_bound or state.f > f_bound:
            continue

        # duplicate detection
        if detect_dupes and visit(visited, state.board):
            continue

        # children
        next_states = get_next_states(state)
        for state in next_states:
            state.g = len(state.history)
            state.f = state.g + weight * heuristic(state.board)
            heapq.heappush(unvisited, state)
        generated += len(next_states)

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def beam(board: Board, **kwargs) -> SearchResult:
    r"""
    Beam search is a variant of breadth-first search that sorts its children greedily
    using a heuristic function and then drops child states to match the beam width.
    This search is incomplete, meaning it may miss a solution although it exists.
    It is useful for limiting memory usage in large search spaces.

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`.
        f_bound (float): A limit on state cost. Default is :math:`\infty`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.
        heuristic: A function that maps boards to an estimated cost-to-go.
            Default is :func:`slidingpuzzle.heuristics.linear_conflict_distance`.
        width (int): The beam width. Default is ``4``.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    f_bound = kwargs.get("f_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)
    heuristic = kwargs.get("heuristic", linear_conflict_distance)
    width = kwargs.get("width", 4)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        next_level = []
        while unvisited:
            state = unvisited.pop()
            expanded += 1

            # goal check
            if np.array_equal(goal, state.board):
                return SearchResult(
                    board, generated, expanded, unvisited, visited, state.history
                )

            # bound
            if len(state.history) > depth_bound or state.f > f_bound:
                continue

            # duplicate detection
            if detect_dupes and visit(visited, state.board):
                continue

            # children
            next_states = get_next_states(state)
            for state in next_states:
                state.f = heuristic(state.board)
            next_level.extend(next_states)
            generated += len(next_states)

        # sort all children at this level
        next_level.sort(reverse=True)
        unvisited = next_level[-width:]

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def bfs(board: Board, **kwargs) -> SearchResult:
    r"""
    Breadth-first search

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    unvisited = collections.deque([initial_state])
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        state = unvisited.popleft()
        expanded += 1

        # goal check
        if np.array_equal(goal, state.board):
            return SearchResult(
                board, generated, expanded, unvisited, visited, state.history
            )

        # bound
        if len(state.history) > depth_bound:
            continue

        # duplicate detection
        if detect_dupes and visit(visited, state.board):
            continue

        # children
        next_states = get_next_states(state)
        unvisited.extend(next_states)
        generated += len(next_states)

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def dfs(board: Board, **kwargs) -> SearchResult:
    r"""
    Depth-first search

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        state = unvisited.pop()
        expanded += 1

        # goal check
        if np.array_equal(goal, state.board):
            return SearchResult(
                board, generated, expanded, unvisited, visited, state.history
            )

        # bound
        if len(state.history) > depth_bound:
            continue

        # duplicate detection
        if detect_dupes and visit(visited, state.board):
            continue

        # children
        next_states = get_next_states(state)
        unvisited.extend(next_states)
        generated += len(next_states)

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def greedy(board: Board, **kwargs) -> SearchResult:
    r"""
    Greedy best-first search. This search orders all known states using the provided
    heuristic and greedily chooses the state closest to the goal.

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`.
        f_bound (float): A limit on state cost. Default is :math:`\infty`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.
        heuristic: A function that maps boards to an estimated cost-to-go.
            Default is :func:`slidingpuzzle.heuristics.linear_conflict_distance`.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    f_bound = kwargs.get("f_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)
    heuristic = kwargs.get("heuristic", linear_conflict_distance)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        state = heapq.heappop(unvisited)
        expanded += 1

        # goal check
        if np.array_equal(goal, state.board):
            return SearchResult(
                board, generated, expanded, unvisited, visited, state.history
            )

        # bound
        if len(state.history) > depth_bound or state.f > f_bound:
            continue

        # duplicate detection
        if detect_dupes and visit(visited, state.board):
            continue

        # children
        next_states = get_next_states(state)
        for state in next_states:
            state.f = heuristic(state.board)
            heapq.heappush(unvisited, state)
        generated += len(next_states)

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def ida_star(board: Board, **kwargs) -> SearchResult:
    r"""
    Iterative deepening A*. A depth-first search that uses an f-bound instead of depth
    to limit search. The next bound is set to the minimum increase in f-bound observed
    during the current iteration. See :func:`a_star`.

    Args:
        board: The board
        depth_bound (int): A limit to search depth. Default is :math:`\infty`. This is
            an optional bound used in addition to the default f-bound.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.
        heuristic: A function that maps boards to an estimated cost-to-go.
            Default is :func:`slidingpuzzle.heuristics.linear_conflict_distance`.
        weight (float): A constant multiplier on heuristic evaluation.
            Default is ``1``.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", float("inf"))
    detect_dupes = kwargs.get("detect_dupes", True)
    heuristic = kwargs.get("heuristic", linear_conflict_distance)
    weight = kwargs.get("weight", 1)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    bound = float(heuristic(board))
    next_bound = float("inf")
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        while unvisited:
            state = unvisited.pop()
            expanded += 1

            # goal check
            if np.array_equal(goal, state.board):
                return SearchResult(
                    board, generated, expanded, unvisited, visited, state.history
                )

            # bound
            if len(state.history) > depth_bound:
                continue
            if state.f > bound:
                next_bound = min(next_bound, state.f)
                continue

            # duplicate detection
            if detect_dupes and visit(visited, state.board):
                continue

            # children
            next_states = get_next_states(state)
            for state in next_states:
                state.g = len(state.history)
                state.f = state.g + weight * heuristic(state.board)
            next_states.sort(reverse=True)
            unvisited.extend(next_states)
            generated += len(next_states)

        # do we need to restart?
        if bound != next_bound and next_bound != float("inf"):
            bound = next_bound
            next_bound = float("inf")
            unvisited.append(initial_state)
            visited.clear()

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def iddfs(board: Board, **kwargs) -> SearchResult:
    r"""
    Iterative deepening depth first search. Similar to :func:`dfs`, except that the
    depth bound is incrementally increased until a solution is found.

    Args:
        board: The board
        depth_bound (int): The initial bound. Default is
            :func:`slidingpuzzle.heuristics.linear_conflict_distance`.
        detect_dupes (bool): Duplicate detection (i.e. track visited states).
            Default is ``True``.

    Returns:
        A :class:`slidingpuzzle.state.SearchResult` with a solution and statistics
    """
    # args
    depth_bound = kwargs.get("depth_bound", linear_conflict_distance(board))
    detect_dupes = kwargs.get("detect_dupes", False)

    # initial state
    goal = new_board(*board.shape)
    initial_state = State(np.copy(board), find_blank(board))
    next_bound = depth_bound
    unvisited = [initial_state]
    visited: set[FrozenBoard] = set()

    # stats
    generated, expanded = 0, 0

    while unvisited:
        while unvisited:
            state = unvisited.pop()
            expanded += 1

            # goal check
            if np.array_equal(goal, state.board):
                return SearchResult(
                    board, generated, expanded, unvisited, visited, state.history
                )

            # bound
            if len(state.history) > depth_bound:
                next_bound = len(state.history)
                continue

            # duplicate detection
            if detect_dupes and visit(visited, state.board):
                continue

            # children
            next_states = get_next_states(state)
            unvisited.extend(next_states)
            generated += len(next_states)

        # do we need to restart?
        if depth_bound != next_bound:
            depth_bound = next_bound
            unvisited.append(initial_state)
            visited.clear()

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


ALGORITHMS_MAP = {
    A_STAR: a_star,
    BEAM: beam,
    BFS: bfs,
    DFS: dfs,
    GREEDY: greedy,
    IDA_STAR: ida_star,
    IDDFS: iddfs,
}


def search(board: Board, alg: str = A_STAR, **kwargs) -> SearchResult:
    r"""
    Searches for a set of moves that take the provided board state to the
    solved state.

    Requested ``alg`` may be one of:
        "a*", "beam", "bfs", "dfs", "greedy", "ida*", "iddfs"

    See :mod:`slidingpuzzle.heuristics` for heuristic functions or provide
    your own.

    If a heuristic is provided with "bfs", "dfs", or "iddfs", it is only used
    to sort the locally generated nodes on each expansion, but not the entire
    open list. You may also provide a "bound" via the kwargs described below
    to limit the search depth.

    If "greedy" or "a*" is used, the entire open list is sorted. The "greedy"
    algorithm sorts only on the heuristic, while "a*" uses heuristic + length
    of the solution. Default heuristics are listed in the table below.

    Of the provided algorithms, only beam search is incomplete. This means it
    may miss the goal, even thought the board is solvable.

    The algorithms support some additional kwargs that can be used to
    customize their behavior. See the docs for individual algorithms.

    Args:
        board: The initial board state to search.
        alg: The algorithm to use for search.
            Use ``print(ALGORITHMS)`` to see available.
        kwargs: Algorithm arguments as described above.

    Returns:
        Returns a :class:`slidingpuzzle.state.SearchResult` containing a list of moves
        to solve the puzzle from the initial state along with some search statistics.
    """

    if not is_solvable(board):
        raise ValueError("The provided board is not solvable.")
    alg_ = alg.strip().lower()
    if alg_ not in ALGORITHMS:
        raise ValueError(f'Unknown algorithm: "{alg}"')
    return ALGORITHMS_MAP[alg_](board, **kwargs)


def evaluate(
    h: int,
    w: int,
    heuristic=linear_conflict_distance,
    alg=A_STAR,
    num_iters: int = 64,
    **kwargs,
) -> float:
    """
    Runs search on ``num_iters`` random boards. Returns the average number of nodes
    generated.

    Args:
        h: Height of the board
        w: Width of the board
        heuristic: Heuristic function to evaluate
        alg: Search algorithm to evaluate
        num_iters: Number of iterations to average
        kwargs: Additional args for ``algorithm``

    Returns:
        The average number of states generated.
    """
    total = 0
    for _ in range(num_iters):
        board = new_board(h, w)
        shuffle_board(board)
        result = search(board, alg=alg, heuristic=heuristic, **kwargs)
        total += result.generated
    return round(total / num_iters, 2)


def compare(
    h: int,
    w: int,
    ha=linear_conflict_distance,
    hb=linear_conflict_distance,
    alga=A_STAR,
    algb=A_STAR,
    num_iters: int = 32,
    kwargsa: Optional[dict] = None,
    kwargsb: Optional[dict] = None,
    **kwargs,
) -> tuple[float, float]:
    """
    Runs search on ``num_iters`` random boards, trying both alga(board, ha) and
    algb(board, hb) on each board. Returns the average number of states generated
    for both.

    Args:
        h: Height of the board
        w: Width of the board
        ha: First heuristic function to evaluate
        hb: Second heuristic function to evaluate
        alga: Search algorithm paired with ``ha``
        algb: Search algorithm paired with ``hb``
        num_iters: Number of iterations to average
        kwargsa: Keyword args for only ``alga``
        kwargsb: Keyword args for only ``algb``
        kwargs: Keyword args for both algorithms

    Returns:
        A tuple containing (avg. generated A, avg. generated B).
    """
    if not kwargsa:
        kwargsa = {}
    if not kwargsb:
        kwargsb = {}
    total_a, total_b = 0, 0
    for _ in range(num_iters):
        board = new_board(h, w)
        shuffle_board(board)
        result = search(board, alg=alga, heuristic=ha, **kwargsa, **kwargs)
        total_a += result.generated
        result = search(board, alg=algb, heuristic=hb, **kwargsb, **kwargs)
        total_b += result.generated
    return round(total_a / num_iters, 2), round(total_b / num_iters, 2)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for dealing with paths required during model training and inference.
"""

from typing import Optional

import pathlib


CHECKPOINT_DIR = "checkpoints"
DATASET_DIR = "datasets"
MODELS_DIR = "models"
TENSORBOARD_DIR = "tensorboard"


def get_board_size_str(h: int, w: int) -> str:
    return f"{h}x{w}"


def get_path(dirname: str, filename: str) -> pathlib.Path:
    """
    Creates intermediate directorys for dirname and returns the full path from dirname
    to filename.

    Args:
        dirname: A directory or path to a directory.
        filename: The filename the path will point to.

    Returns:
        The path object.
    """
    dirpath = pathlib.Path(dirname)
    dirpath.mkdir(exist_ok=True, parents=True)
    return dirpath / filename


def get_checkpoint_path(h: int, w: int, tag: str) -> pathlib.Path:
    """
    Get the path to a checkpoint, given a board size and optional tag.

    Args:
        h: Board height
        w: Board width
        tag: Checkpoint tag to load
    """
    checkpoint_name = f"{get_board_size_str(h, w)}_{tag}"
    return get_path(CHECKPOINT_DIR, f"checkpoint_{checkpoint_name}")


def get_examples_path(h: int, w: int) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(DATASET_DIR, f"examples_{board_size_str}.json")


def get_model_path(h: int, w: int, version: str) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(MODELS_DIR, f"{version}_{board_size_str}.pt")


def get_log_dir(tensorboard_dir: str, h: int, w: int) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(tensorboard_dir, f"slidingpuzzle_{board_size_str}")
